@@ -40,36 +40,19 @@
 #include "src/dbus.h"
 #include "src/knownnetworks.h"
 #include "src/scan.h"
+#include "src/util.h"
 
 static struct l_queue *known_networks;
 static size_t num_known_hidden_networks;
 static struct l_dir_watch *storage_dir_watch;
 
-static int timespec_compare(const struct timespec *tsa,
-				const struct timespec *tsb)
-{
-
-	if (tsa->tv_sec > tsb->tv_sec)
-		return -1;
-
-	if (tsa->tv_sec < tsb->tv_sec)
-		return 1;
-
-	if (tsa->tv_nsec > tsb->tv_nsec)
-		return -1;
-
-	if (tsa->tv_nsec < tsb->tv_nsec)
-		return -1;
-
-	return 0;
-}
-
 static int connected_time_compare(const void *a, const void *b, void *user_data)
 {
 	const struct network_info *ni_a = a;
 	const struct network_info *ni_b = b;
 
-	return timespec_compare(&ni_a->connected_time, &ni_b->connected_time);
+	return util_timespec_compare(&ni_a->connected_time,
+					&ni_b->connected_time);
 }
 
 const char *known_network_get_path(const struct network_info *network)
@@ -131,7 +114,7 @@ static void known_network_update(struct network_info *orig_network,
 	else
 		network = network_info_add_known(ssid, security);
 
-	if (timespec_compare(&network->connected_time, connected_time) &&
+	if (util_timespec_compare(&network->connected_time, connected_time) &&
 			orig_network) {
 		l_dbus_property_changed(dbus_get_bus(),
 					known_network_get_path(network),
@@ -605,7 +588,7 @@ static void known_network_frequencies_sync(void)
 	l_settings_free(known_freqs);
 }
 
-bool known_networks_init(void)
+static int known_networks_init(void)
 {
 	struct l_dbus *dbus = dbus_get_bus();
 	DIR *dir;
@@ -616,7 +599,7 @@ bool known_networks_init(void)
 						NULL, false)) {
 		l_info("Unable to register %s interface",
 				IWD_KNOWN_NETWORK_INTERFACE);
-		return false;
+		return -EPERM;
 	}
 
 	dir = opendir(DAEMON_STORAGEDIR);
@@ -624,7 +607,7 @@ bool known_networks_init(void)
 		l_info("Unable to open %s: %s", DAEMON_STORAGEDIR,
 							strerror(errno));
 		l_dbus_unregister_interface(dbus, IWD_KNOWN_NETWORK_INTERFACE);
-		return false;
+		return -ENOENT;
 	}
 
 	known_networks = l_queue_new();
@@ -661,10 +644,10 @@ bool known_networks_init(void)
 						known_networks_watch_cb, NULL,
 						known_networks_watch_destroy);
 
-	return true;
+	return 0;
 }
 
-void known_networks_exit(void)
+static void known_networks_exit(void)
 {
 	struct l_dbus *dbus = dbus_get_bus();
 
@@ -677,3 +660,5 @@ void known_networks_exit(void)
 
 	l_dbus_unregister_interface(dbus, IWD_KNOWN_NETWORK_INTERFACE);
 }
+
+IWD_MODULE(known_networks, known_networks_init, known_networks_exit)
