@@ -43,12 +43,6 @@
 #include "src/common.h"
 #include "src/storage.h"
 
-#ifdef TEMP_FAILURE_RETRY
-#define TFR TEMP_FAILURE_RETRY
-#else
-#define TFR
-#endif
-
 #define STORAGE_DIR_MODE (S_IRUSR | S_IWUSR | S_IXUSR)
 #define STORAGE_FILE_MODE (S_IRUSR | S_IWUSR)
 
@@ -97,16 +91,16 @@ ssize_t read_file(void *buffer, size_t len, const char *path_fmt, ...)
 	path = l_strdup_vprintf(path_fmt, ap);
 	va_end(ap);
 
-	fd = TFR(open(path, O_RDONLY));
+	fd = L_TFR(open(path, O_RDONLY));
 
 	l_free(path);
 
 	if (fd == -1)
 		return -1;
 
-	r = TFR(read(fd, buffer, len));
+	r = L_TFR(read(fd, buffer, len));
 
-	TFR(close(fd));
+	L_TFR(close(fd));
 
 	return r;
 }
@@ -139,12 +133,12 @@ ssize_t write_file(const void *buffer, size_t len,
 	if (create_dirs(path) != 0)
 		goto error_create_dirs;
 
-	fd = TFR(mkostemps(tmp_path, 4, O_CLOEXEC));
+	fd = L_TFR(mkostemps(tmp_path, 4, O_CLOEXEC));
 	if (fd == -1)
 		goto error_mkostemps;
 
-	r = TFR(write(fd, buffer, len));
-	TFR(close(fd));
+	r = L_TFR(write(fd, buffer, len));
+	L_TFR(close(fd));
 
 	if (r != (ssize_t) len) {
 		r = -1;
@@ -343,3 +337,33 @@ int storage_network_remove(enum security type, const char *ssid)
 
 	return ret < 0 ? -errno : 0;
 }
+
+static const char *known_freq_file_path =
+				DAEMON_STORAGEDIR "/.known_network.freq";
+
+struct l_settings *storage_known_frequencies_load(void)
+{
+	struct l_settings *known_freqs;
+
+	known_freqs = l_settings_new();
+
+	if (!l_settings_load_from_file(known_freqs, known_freq_file_path)) {
+		l_settings_free(known_freqs);
+		known_freqs = NULL;
+	}
+
+	return known_freqs;
+}
+
+void storage_known_frequencies_sync(struct l_settings *known_freqs)
+{
+	char *data;
+	size_t len;
+
+	if (!known_freqs)
+		return;
+
+	data = l_settings_to_data(known_freqs, &len);
+	write_file(data, len, "%s", known_freq_file_path);
+	l_free(data);
+}
